from glob import glob
from io import StringIO
from os.path import basename
from os.path import join as p_join

import numpy as np
import pandas as pd

from openqdc.datasets.base import BaseDataset
from openqdc.methods import PotentialMethod
from openqdc.utils.constants import ATOM_TABLE


def read_xyz_and_get_content(fname):
    with open(fname, "r") as f:
        content = f.read()
    s = StringIO(content)
    d = np.loadtxt(s, skiprows=2, dtype="str")
    z, positions = d[:, 0], d[:, 1:].astype(np.float32)
    z = np.array([ATOM_TABLE.GetAtomicNumber(s) for s in z])
    xs = np.stack((z, np.zeros_like(z)), axis=-1)
    conf = dict(
        atomic_inputs=np.concatenate((xs, positions), axis=-1, dtype=np.float32),
        name=np.array([basename(fname)]),
        n_atoms=np.array([positions.shape[0]], dtype=np.int32),
        subset=np.array(["multixcqm9"]),
    )
    return conf


def read_xyz_files(folder_path):
    xyz_list = []
    for file in sorted(glob(folder_path + "/xyzfiles/*.xyz")):
        xyz_list.append(read_xyz_and_get_content(file))
    return xyz_list


class MultixcQM9(BaseDataset):
    """
    MultixcQM9 is a dataset of molecular and reaction energies from multi-level quantum chemical methods consisting
    of 133K QM9 molecules geometries calculated with 76 different DFT functionals and three different basis sets
    resulting in 228 energy values for each molecule along with semi-empirical method GFN2-xTB. Geometries for the
    molecules are used directly from Kim et al. which uses G4MP2 method.

    Usage:
    ```python
    from openqdc.datasets import MultixcQM9
    dataset = MultixcQM9()
    ```

    References:
        https://www.nature.com/articles/s41597-023-02690-2\n
        https://github.com/chemsurajit/largeDFTdata\n
        https://www.nature.com/articles/s41597-019-0121-7\n
    """

    __name__ = "multixcqm9"

    __energy_methods__ = [
        PotentialMethod.KCIS_MODIFIED_DZP,
        PotentialMethod.KCIS_ORIGINAL_DZP,
        PotentialMethod.PKZB_DZP,
        PotentialMethod.VS98_DZP,
        PotentialMethod.LDA_VWN_DZP,
        PotentialMethod.PW91_DZP,
        PotentialMethod.BLYP_DZP,
        PotentialMethod.BP_DZP,
        PotentialMethod.PBE_DZP,
        PotentialMethod.RPBE_DZP,
        PotentialMethod.REVPBE_DZP,
        PotentialMethod.OLYP_DZP,
        PotentialMethod.FT97_DZP,
        PotentialMethod.BLAP3_DZP,
        PotentialMethod.HCTH_93_DZP,
        PotentialMethod.HCTH_120_DZP,
        PotentialMethod.HCTH_147_DZP,
        PotentialMethod.HCTH_407_DZP,
        PotentialMethod.BMTAU1_DZP,
        PotentialMethod.BOP_DZP,
        PotentialMethod.PKZBX_KCISCOR_DZP,
        PotentialMethod.VS98_X_XC_DZP,
        PotentialMethod.VS98_X_ONLY_DZP,
        PotentialMethod.BECKE00_DZP,
        PotentialMethod.BECKE00X_XC_DZP,
        PotentialMethod.BECKE00_X_ONLY_DZP,
        PotentialMethod.BECKE88X_BR89C_DZP,
        PotentialMethod.OLAP3_DZP,
        PotentialMethod.TPSS_DZP,
        PotentialMethod.MPBE_DZP,
        PotentialMethod.OPBE_DZP,
        PotentialMethod.OPERDEW_DZP,
        PotentialMethod.MPBEKCIS_DZP,
        PotentialMethod.MPW_DZP,
        PotentialMethod.TAU_HCTH_DZP,
        PotentialMethod.XLYP_DZP,
        PotentialMethod.KT1_DZP,
        PotentialMethod.KT2_DZP,
        PotentialMethod.M06_L_DZP,
        PotentialMethod.BLYP_D_DZP,
        PotentialMethod.BP86_D_DZP,
        PotentialMethod.PBE_D_DZP,
        PotentialMethod.TPSSD_DZP,
        PotentialMethod.B97_D_DZP,
        PotentialMethod.REVTPSS_DZP,
        PotentialMethod.PBESOL_DZP,
        PotentialMethod.RGE2_DZP,
        PotentialMethod.SSB_D_DZP,
        PotentialMethod.MVS_DZP,
        PotentialMethod.MVSX_DZP,
        PotentialMethod.TMGGA_DZP,
        PotentialMethod.TPSSH_DZP,
        PotentialMethod.B3LYP_VWN5_DZP,
        PotentialMethod.O3LYP_VWN5_DZP,
        PotentialMethod.KMLYP_VWN5_DZP,
        PotentialMethod.PBE0_DZP,
        PotentialMethod.B3LYP_S_VWN5_DZP,
        PotentialMethod.BHANDH_DZP,
        PotentialMethod.BHANDHLYP_DZP,
        PotentialMethod.B97_DZP,
        PotentialMethod.B97_1_DZP,
        PotentialMethod.B97_2_DZP,
        PotentialMethod.MPBE0KCIS_DZP,
        PotentialMethod.MPBE1KCIS_DZP,
        PotentialMethod.B1LYP_VWN5_DZP,
        PotentialMethod.B1PW91_VWN5_DZP,
        PotentialMethod.MPW1PW_DZP,
        PotentialMethod.MPW1K_DZP,
        PotentialMethod.TAU_HCTH_HYBRID_DZP,
        PotentialMethod.X3LYP_VWN5_DZP,
        PotentialMethod.OPBE0_DZP,
        PotentialMethod.M05_DZP,
        PotentialMethod.M05_2X_DZP,
        PotentialMethod.M06_DZP,
        PotentialMethod.M06_2X_DZP,
        PotentialMethod.B3LYP_D_DZP,
        PotentialMethod.KCIS_MODIFIED_TZP,
        PotentialMethod.KCIS_ORIGINAL_TZP,
        PotentialMethod.PKZB_TZP,
        PotentialMethod.VS98_TZP,
        PotentialMethod.LDA_VWN_TZP,
        PotentialMethod.PW91_TZP,
        PotentialMethod.BLYP_TZP,
        PotentialMethod.BP_TZP,
        PotentialMethod.PBE_TZP,
        PotentialMethod.RPBE_TZP,
        PotentialMethod.REVPBE_TZP,
        PotentialMethod.OLYP_TZP,
        PotentialMethod.FT97_TZP,
        PotentialMethod.BLAP3_TZP,
        PotentialMethod.HCTH_93_TZP,
        PotentialMethod.HCTH_120_TZP,
        PotentialMethod.HCTH_147_TZP,
        PotentialMethod.HCTH_407_TZP,
        PotentialMethod.BMTAU1_TZP,
        PotentialMethod.BOP_TZP,
        PotentialMethod.PKZBX_KCISCOR_TZP,
        PotentialMethod.VS98_X_XC_TZP,
        PotentialMethod.VS98_X_ONLY_TZP,
        PotentialMethod.BECKE00_TZP,
        PotentialMethod.BECKE00X_XC_TZP,
        PotentialMethod.BECKE00_X_ONLY_TZP,
        PotentialMethod.BECKE88X_BR89C_TZP,
        PotentialMethod.OLAP3_TZP,
        PotentialMethod.TPSS_TZP,
        PotentialMethod.MPBE_TZP,
        PotentialMethod.OPBE_TZP,
        PotentialMethod.OPERDEW_TZP,
        PotentialMethod.MPBEKCIS_TZP,
        PotentialMethod.MPW_TZP,
        PotentialMethod.TAU_HCTH_TZP,
        PotentialMethod.XLYP_TZP,
        PotentialMethod.KT1_TZP,
        PotentialMethod.KT2_TZP,
        PotentialMethod.M06_L_TZP,
        PotentialMethod.BLYP_D_TZP,
        PotentialMethod.BP86_D_TZP,
        PotentialMethod.PBE_D_TZP,
        PotentialMethod.TPSSD_TZP,
        PotentialMethod.B97_D_TZP,
        PotentialMethod.REVTPSS_TZP,
        PotentialMethod.PBESOL_TZP,
        PotentialMethod.RGE2_TZP,
        PotentialMethod.SSB_D_TZP,
        PotentialMethod.MVS_TZP,
        PotentialMethod.MVSX_TZP,
        PotentialMethod.TMGGA_TZP,
        PotentialMethod.TPSSH_TZP,
        PotentialMethod.B3LYP_VWN5_TZP,
        PotentialMethod.O3LYP_VWN5_TZP,
        PotentialMethod.KMLYP_VWN5_TZP,
        PotentialMethod.PBE0_TZP,
        PotentialMethod.B3LYP_S_VWN5_TZP,
        PotentialMethod.BHANDH_TZP,
        PotentialMethod.BHANDHLYP_TZP,
        PotentialMethod.B97_TZP,
        PotentialMethod.B97_1_TZP,
        PotentialMethod.B97_2_TZP,
        PotentialMethod.MPBE0KCIS_TZP,
        PotentialMethod.MPBE1KCIS_TZP,
        PotentialMethod.B1LYP_VWN5_TZP,
        PotentialMethod.B1PW91_VWN5_TZP,
        PotentialMethod.MPW1PW_TZP,
        PotentialMethod.MPW1K_TZP,
        PotentialMethod.TAU_HCTH_HYBRID_TZP,
        PotentialMethod.X3LYP_VWN5_TZP,
        PotentialMethod.OPBE0_TZP,
        PotentialMethod.M05_TZP,
        PotentialMethod.M05_2X_TZP,
        PotentialMethod.M06_TZP,
        PotentialMethod.M06_2X_TZP,
        PotentialMethod.B3LYP_D_TZP,
        PotentialMethod.KCIS_MODIFIED_SZ,
        PotentialMethod.KCIS_ORIGINAL_SZ,
        PotentialMethod.PKZB_SZ,
        PotentialMethod.VS98_SZ,
        PotentialMethod.LDA_VWN_SZ,
        PotentialMethod.PW91_SZ,
        PotentialMethod.BLYP_SZ,
        PotentialMethod.BP_SZ,
        PotentialMethod.PBE_SZ,
        PotentialMethod.RPBE_SZ,
        PotentialMethod.REVPBE_SZ,
        PotentialMethod.OLYP_SZ,
        PotentialMethod.FT97_SZ,
        PotentialMethod.BLAP3_SZ,
        PotentialMethod.HCTH_93_SZ,
        PotentialMethod.HCTH_120_SZ,
        PotentialMethod.HCTH_147_SZ,
        PotentialMethod.HCTH_407_SZ,
        PotentialMethod.BMTAU1_SZ,
        PotentialMethod.BOP_SZ,
        PotentialMethod.PKZBX_KCISCOR_SZ,
        PotentialMethod.VS98_X_XC_SZ,
        PotentialMethod.VS98_X_ONLY_SZ,
        PotentialMethod.BECKE00_SZ,
        PotentialMethod.BECKE00X_XC_SZ,
        PotentialMethod.BECKE00_X_ONLY_SZ,
        PotentialMethod.BECKE88X_BR89C_SZ,
        PotentialMethod.OLAP3_SZ,
        PotentialMethod.TPSS_SZ,
        PotentialMethod.MPBE_SZ,
        PotentialMethod.OPBE_SZ,
        PotentialMethod.OPERDEW_SZ,
        PotentialMethod.MPBEKCIS_SZ,
        PotentialMethod.MPW_SZ,
        PotentialMethod.TAU_HCTH_SZ,
        PotentialMethod.XLYP_SZ,
        PotentialMethod.KT1_SZ,
        PotentialMethod.KT2_SZ,
        PotentialMethod.M06_L_SZ,
        PotentialMethod.BLYP_D_SZ,
        PotentialMethod.BP86_D_SZ,
        PotentialMethod.PBE_D_SZ,
        PotentialMethod.TPSSD_SZ,
        PotentialMethod.B97_D_SZ,
        PotentialMethod.REVTPSS_SZ,
        PotentialMethod.PBESOL_SZ,
        PotentialMethod.RGE2_SZ,
        PotentialMethod.SSB_D_SZ,
        PotentialMethod.MVS_SZ,
        PotentialMethod.MVSX_SZ,
        PotentialMethod.TMGGA_SZ,
        PotentialMethod.TPSSH_SZ,
        PotentialMethod.B3LYP_VWN5_SZ,
        PotentialMethod.O3LYP_VWN5_SZ,
        PotentialMethod.KMLYP_VWN5_SZ,
        PotentialMethod.PBE0_SZ,
        PotentialMethod.B3LYP_S_VWN5_SZ,
        PotentialMethod.BHANDH_SZ,
        PotentialMethod.BHANDHLYP_SZ,
        PotentialMethod.B97_SZ,
        PotentialMethod.B97_1_SZ,
        PotentialMethod.B97_2_SZ,
        PotentialMethod.MPBE0KCIS_SZ,
        PotentialMethod.MPBE1KCIS_SZ,
        PotentialMethod.B1LYP_VWN5_SZ,
        PotentialMethod.B1PW91_VWN5_SZ,
        PotentialMethod.MPW1PW_SZ,
        PotentialMethod.MPW1K_SZ,
        PotentialMethod.TAU_HCTH_HYBRID_SZ,
        PotentialMethod.X3LYP_VWN5_SZ,
        PotentialMethod.OPBE0_SZ,
        PotentialMethod.M05_SZ,
        PotentialMethod.M05_2X_SZ,
        PotentialMethod.M06_SZ,
        PotentialMethod.M06_2X_SZ,
        PotentialMethod.B3LYP_D_SZ,
        PotentialMethod.GFN2_XTB,
    ]

    energy_target_names = [
        "KCIS-MODIFIED/DZP",
        "KCIS-ORIGINAL/DZP",
        "PKZB/DZP",
        "VS98/DZP",
        "LDA(VWN)/DZP",
        "PW91/DZP",
        "BLYP/DZP",
        "BP/DZP",
        "PBE/DZP",
        "RPBE/DZP",
        "REVPBE/DZP",
        "OLYP/DZP",
        "FT97/DZP",
        "BLAP3/DZP",
        "HCTH/93/DZP",
        "HCTH/120/DZP",
        "HCTH/147/DZP",
        "HCTH/407/DZP",
        "BMTAU1/DZP",
        "BOP/DZP",
        "PKZBX-KCISCOR/DZP",
        "VS98-X(XC)/DZP",
        "VS98-X-ONLY/DZP",
        "BECKE00/DZP",
        "BECKE00X(XC)/DZP",
        "BECKE00-X-ONLY/DZP",
        "BECKE88X+BR89C/DZP",
        "OLAP3/DZP",
        "TPSS/DZP",
        "MPBE/DZP",
        "OPBE/DZP",
        "OPERDEW/DZP",
        "MPBEKCIS/DZP",
        "MPW/DZP",
        "TAU-HCTH/DZP",
        "XLYP/DZP",
        "KT1/DZP",
        "KT2/DZP",
        "M06-L/DZP",
        "BLYP-D/DZP",
        "BP86-D/DZP",
        "PBE-D/DZP",
        "TPSS-D/DZP",
        "B97-D/DZP",
        "REVTPSS/DZP",
        "PBESOL/DZP",
        "RGE2/DZP",
        "SSB-D/DZP",
        "MVS/DZP",
        "MVSX/DZP",
        "T-MGGA/DZP",
        "TPSSH/DZP",
        "B3LYP(VWN5)/DZP",
        "O3LYP(VWN5)/DZP",
        "KMLYP(VWN5)/DZP",
        "PBE0/DZP",
        "B3LYP*(VWN5)/DZP",
        "BHANDH/DZP",
        "BHANDHLYP/DZP",
        "B97/DZP",
        "B97-1/DZP",
        "B97-2/DZP",
        "MPBE0KCIS/DZP",
        "MPBE1KCIS/DZP",
        "B1LYP(VWN5)/DZP",
        "B1PW91(VWN5)/DZP",
        "MPW1PW/DZP",
        "MPW1K/DZP",
        "TAU-HCTH-HYBRID/DZP",
        "X3LYP(VWN5)/DZP",
        "OPBE0/DZP",
        "M05/DZP",
        "M05-2X/DZP",
        "M06/DZP",
        "M06-2X/DZP",
        "B3LYP-D/DZP",
        "KCIS-MODIFIED/TZP",
        "KCIS-ORIGINAL/TZP",
        "PKZB/TZP",
        "VS98/TZP",
        "LDA(VWN)/TZP",
        "PW91/TZP",
        "BLYP/TZP",
        "BP/TZP",
        "PBE/TZP",
        "RPBE/TZP",
        "REVPBE/TZP",
        "OLYP/TZP",
        "FT97/TZP",
        "BLAP3/TZP",
        "HCTH/93/TZP",
        "HCTH/120/TZP",
        "HCTH/147/TZP",
        "HCTH/407/TZP",
        "BMTAU1/TZP",
        "BOP/TZP",
        "PKZBX-KCISCOR/TZP",
        "VS98-X(XC)/TZP",
        "VS98-X-ONLY/TZP",
        "BECKE00/TZP",
        "BECKE00X(XC)/TZP",
        "BECKE00-X-ONLY/TZP",
        "BECKE88X+BR89C/TZP",
        "OLAP3/TZP",
        "TPSS/TZP",
        "MPBE/TZP",
        "OPBE/TZP",
        "OPERDEW/TZP",
        "MPBEKCIS/TZP",
        "MPW/TZP",
        "TAU-HCTH/TZP",
        "XLYP/TZP",
        "KT1/TZP",
        "KT2/TZP",
        "M06-L/TZP",
        "BLYP-D/TZP",
        "BP86-D/TZP",
        "PBE-D/TZP",
        "TPSS-D/TZP",
        "B97-D/TZP",
        "REVTPSS/TZP",
        "PBESOL/TZP",
        "RGE2/TZP",
        "SSB-D/TZP",
        "MVS/TZP",
        "MVSX/TZP",
        "T-MGGA/TZP",
        "TPSSH/TZP",
        "B3LYP(VWN5)/TZP",
        "O3LYP(VWN5)/TZP",
        "KMLYP(VWN5)/TZP",
        "PBE0/TZP",
        "B3LYP*(VWN5)/TZP",
        "BHANDH/TZP",
        "BHANDHLYP/TZP",
        "B97/TZP",
        "B97-1/TZP",
        "B97-2/TZP",
        "MPBE0KCIS/TZP",
        "MPBE1KCIS/TZP",
        "B1LYP(VWN5)/TZP",
        "B1PW91(VWN5)/TZP",
        "MPW1PW/TZP",
        "MPW1K/TZP",
        "TAU-HCTH-HYBRID/TZP",
        "X3LYP(VWN5)/TZP",
        "OPBE0/TZP",
        "M05/TZP",
        "M05-2X/TZP",
        "M06/TZP",
        "M06-2X/TZP",
        "B3LYP-D/TZP",
        "KCIS-MODIFIED/SZ",
        "KCIS-ORIGINAL/SZ",
        "PKZB/SZ",
        "VS98/SZ",
        "LDA(VWN)/SZ",
        "PW91/SZ",
        "BLYP/SZ",
        "BP/SZ",
        "PBE/SZ",
        "RPBE/SZ",
        "REVPBE/SZ",
        "OLYP/SZ",
        "FT97/SZ",
        "BLAP3/SZ",
        "HCTH/93/SZ",
        "HCTH/120/SZ",
        "HCTH/147/SZ",
        "HCTH/407/SZ",
        "BMTAU1/SZ",
        "BOP/SZ",
        "PKZBX-KCISCOR/SZ",
        "VS98-X(XC)/SZ",
        "VS98-X-ONLY/SZ",
        "BECKE00/SZ",
        "BECKE00X(XC)/SZ",
        "BECKE00-X-ONLY/SZ",
        "BECKE88X+BR89C/SZ",
        "OLAP3/SZ",
        "TPSS/SZ",
        "MPBE/SZ",
        "OPBE/SZ",
        "OPERDEW/SZ",
        "MPBEKCIS/SZ",
        "MPW/SZ",
        "TAU-HCTH/SZ",
        "XLYP/SZ",
        "KT1/SZ",
        "KT2/SZ",
        "M06-L/SZ",
        "BLYP-D/SZ",
        "BP86-D/SZ",
        "PBE-D/SZ",
        "TPSS-D/SZ",
        "B97-D/SZ",
        "REVTPSS/SZ",
        "PBESOL/SZ",
        "RGE2/SZ",
        "SSB-D/SZ",
        "MVS/SZ",
        "MVSX/SZ",
        "T-MGGA/SZ",
        "TPSSH/SZ",
        "B3LYP(VWN5)/SZ",
        "O3LYP(VWN5)/SZ",
        "KMLYP(VWN5)/SZ",
        "PBE0/SZ",
        "B3LYP*(VWN5)/SZ",
        "BHANDH/SZ",
        "BHANDHLYP/SZ",
        "B97/SZ",
        "B97-1/SZ",
        "B97-2/SZ",
        "MPBE0KCIS/SZ",
        "MPBE1KCIS/SZ",
        "B1LYP(VWN5)/SZ",
        "B1PW91(VWN5)/SZ",
        "MPW1PW/SZ",
        "MPW1K/SZ",
        "TAU-HCTH-HYBRID/SZ",
        "X3LYP(VWN5)/SZ",
        "OPBE0/SZ",
        "M05/SZ",
        "M05-2X/SZ",
        "M06/SZ",
        "M06-2X/SZ",
        "B3LYP-D/SZ",
        "GFNXTB",
    ]

    __energy_unit__ = "ev"  # to fix
    __distance_unit__ = "ang"  # to fix
    __forces_unit__ = "ev/ang"  # to fix
    __links__ = {
        "xyz.zip": "https://data.dtu.dk/ndownloader/files/35143624",
        "xtb.zip": "https://data.dtu.dk/ndownloader/files/42444300",
        "dzp.zip": "https://data.dtu.dk/ndownloader/files/42443925",
        "tzp.zip": "https://data.dtu.dk/ndownloader/files/42444129",
        "sz.zip": "https://data.dtu.dk/ndownloader/files/42441345",
        "failed_indices.dat": "https://data.dtu.dk/ndownloader/files/37337677",
    }

    def _read_molecules_energies(self):
        d = {"DZP": None, "TZP": None, "SZ": None, "XTB": None}
        for basis in d.keys():
            d[basis] = pd.read_csv(p_join(self.root, basis, "molecules/molecules.csv"), index_col=False).drop(
                columns=["index"]
            )
        return pd.concat([d["DZP"], d["TZP"], d["SZ"], d["XTB"]], axis=1, ignore_index=False)

    def _read_all_xyzs(self):
        xyz_list = read_xyz_files(self.root)
        return pd.DataFrame(xyz_list)

    def read_raw_entries(self):
        df_energies = self._read_molecules_energies()
        df_xyz = self._read_all_xyzs()
        return [
            {"energies": np.atleast_2d(en), **xyz_dict}
            for xyz_dict, en in zip(df_xyz.to_dict("records"), df_energies.values.astype(np.float64))
        ]


class MultixcQM9_V2(MultixcQM9):
    __name__ = "multixcqm9_v2"

    __energy_methods__ = MultixcQM9.__energy_methods__ + [PotentialMethod.PM6]
    energy_target_names = MultixcQM9.energy_target_names + ["PM6"]
    __force_mask__ = MultixcQM9.__force_mask__ + [False]
